# Generated from QueryExprParser.g4 by ANTLR 4.12.0
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,16,33,2,0,7,0,2,1,7,1,1,0,5,0,6,8,0,10,0,12,0,9,9,0,1,0,1,0,
        1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,3,1,
        28,8,1,1,1,3,1,31,8,1,1,1,0,0,2,0,2,0,0,36,0,7,1,0,0,0,2,27,1,0,
        0,0,4,6,3,2,1,0,5,4,1,0,0,0,6,9,1,0,0,0,7,5,1,0,0,0,7,8,1,0,0,0,
        8,10,1,0,0,0,9,7,1,0,0,0,10,11,5,0,0,1,11,1,1,0,0,0,12,13,5,1,0,
        0,13,14,5,3,0,0,14,28,5,15,0,0,15,16,5,1,0,0,16,17,5,3,0,0,17,28,
        5,16,0,0,18,19,5,1,0,0,19,20,5,3,0,0,20,28,5,12,0,0,21,22,5,1,0,
        0,22,23,5,3,0,0,23,28,5,13,0,0,24,25,5,1,0,0,25,26,5,3,0,0,26,28,
        5,14,0,0,27,12,1,0,0,0,27,15,1,0,0,0,27,18,1,0,0,0,27,21,1,0,0,0,
        27,24,1,0,0,0,28,30,1,0,0,0,29,31,5,10,0,0,30,29,1,0,0,0,30,31,1,
        0,0,0,31,3,1,0,0,0,3,7,27,30
    ]

class QueryExprParser ( Parser ):

    grammarFileName = "QueryExprParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "'!='", "'>'", "'>='", "'<'", "'<='", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "'NULL'" ]

    symbolicNames = [ "<INVALID>", "IDENTIFIER", "WS", "OPERATOR", "EQ", 
                      "NE", "GT", "GE", "LT", "LE", "LABEL", "DELIMITER", 
                      "NUMBER", "BOOL", "NULL", "STRING", "STRING_QUOTED" ]

    RULE_exprs = 0
    RULE_expr = 1

    ruleNames =  [ "exprs", "expr" ]

    EOF = Token.EOF
    IDENTIFIER=1
    WS=2
    OPERATOR=3
    EQ=4
    NE=5
    GT=6
    GE=7
    LT=8
    LE=9
    LABEL=10
    DELIMITER=11
    NUMBER=12
    BOOL=13
    NULL=14
    STRING=15
    STRING_QUOTED=16

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.12.0")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class ExprsContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(QueryExprParser.EOF, 0)

        def expr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(QueryExprParser.ExprContext)
            else:
                return self.getTypedRuleContext(QueryExprParser.ExprContext,i)


        def getRuleIndex(self):
            return QueryExprParser.RULE_exprs

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExprs" ):
                listener.enterExprs(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExprs" ):
                listener.exitExprs(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitExprs" ):
                return visitor.visitExprs(self)
            else:
                return visitor.visitChildren(self)




    def exprs(self):

        localctx = QueryExprParser.ExprsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_exprs)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 7
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==1:
                self.state = 4
                self.expr()
                self.state = 9
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 10
            self.match(QueryExprParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ExprContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(QueryExprParser.IDENTIFIER, 0)

        def OPERATOR(self):
            return self.getToken(QueryExprParser.OPERATOR, 0)

        def STRING(self):
            return self.getToken(QueryExprParser.STRING, 0)

        def STRING_QUOTED(self):
            return self.getToken(QueryExprParser.STRING_QUOTED, 0)

        def NUMBER(self):
            return self.getToken(QueryExprParser.NUMBER, 0)

        def BOOL(self):
            return self.getToken(QueryExprParser.BOOL, 0)

        def NULL(self):
            return self.getToken(QueryExprParser.NULL, 0)

        def LABEL(self):
            return self.getToken(QueryExprParser.LABEL, 0)

        def getRuleIndex(self):
            return QueryExprParser.RULE_expr

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterExpr" ):
                listener.enterExpr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitExpr" ):
                listener.exitExpr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitExpr" ):
                return visitor.visitExpr(self)
            else:
                return visitor.visitChildren(self)




    def expr(self):

        localctx = QueryExprParser.ExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_expr)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 27
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,1,self._ctx)
            if la_ == 1:
                self.state = 12
                self.match(QueryExprParser.IDENTIFIER)
                self.state = 13
                self.match(QueryExprParser.OPERATOR)
                self.state = 14
                self.match(QueryExprParser.STRING)
                pass

            elif la_ == 2:
                self.state = 15
                self.match(QueryExprParser.IDENTIFIER)
                self.state = 16
                self.match(QueryExprParser.OPERATOR)
                self.state = 17
                self.match(QueryExprParser.STRING_QUOTED)
                pass

            elif la_ == 3:
                self.state = 18
                self.match(QueryExprParser.IDENTIFIER)
                self.state = 19
                self.match(QueryExprParser.OPERATOR)
                self.state = 20
                self.match(QueryExprParser.NUMBER)
                pass

            elif la_ == 4:
                self.state = 21
                self.match(QueryExprParser.IDENTIFIER)
                self.state = 22
                self.match(QueryExprParser.OPERATOR)
                self.state = 23
                self.match(QueryExprParser.BOOL)
                pass

            elif la_ == 5:
                self.state = 24
                self.match(QueryExprParser.IDENTIFIER)
                self.state = 25
                self.match(QueryExprParser.OPERATOR)
                self.state = 26
                self.match(QueryExprParser.NULL)
                pass


            self.state = 30
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==10:
                self.state = 29
                self.match(QueryExprParser.LABEL)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence

from antlr4 import CommonTokenStream, InputStream

from .generated.QueryExprLexer import QueryExprLexer
from .generated.QueryExprParser import QueryExprParser
from .generated.QueryExprParserVisitor import QueryExprParserVisitor


@dataclass(frozen=True, eq=True)
class Filter:
    """Represents a filter construct.

    Filters are immutable, and supports equality checks, eg.
    >>> Filter("foo", "=", "bar") == Filter("foo", "=", "bar")
    True

    """

    identifier: str
    operator: str = "="
    value: Any = ""
    label: Optional[str] = None

    @property
    def value_str(self) -> str:
        """Returns the string representation of `value`.
        * None   : "NULL"
        * boolean: "true" or "false"
        * number : "123" or "1.23"
        * string : as-is
        """
        if isinstance(self.value, str):
            return self.value
        elif isinstance(self.value, bool):
            return "true" if self.value else "false"
        elif self.value is None:
            return "NULL"
        else:
            return str(self.value)

    def __repr__(self) -> str:
        return f"<Filter {self.to_string()}>"

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        """Return a filter expression.

        >>> f = Filter("foo", "=", "bar")
        >>> f.to_string()
        "foo=bar"
        """
        parts = [self.identifier or "", self.operator or ""]
        parts.append(safe_string(self.value_str))
        if self.label:
            parts.append(f"[{safe_string(self.label)}]")
        return "".join(parts)

    def to_json(self) -> Dict:
        """Returns a dictionary representation.
        >>> f = Filter("foo", "=", "bar")
        >>> f.to_json()
        {
            "identifier": "foo",
            "operator": "=",
            "value": "bar"
        }
        """
        obj = {
            "identifier": self.identifier,
            "operator": self.operator,
            "value": self.value,
        }
        if self.label:
            obj["label"] = self.label
        return obj


class Query:
    """Query class parses and generates query expression."""

    @classmethod
    def parse(cls: Any, query: str) -> Query:
        """Return a `Query` by parsing the `query` expression.

        >>> q = Query.parse("foo=bar foo=baz")
        >>> q.filters
        [<Filter foo=bar>, <Filter foo=baz>]

        """
        lexer = QueryExprLexer(InputStream(query))
        parser = QueryExprParser(CommonTokenStream(lexer))
        filters = _Visitor().visitExprs(parser.exprs())
        return cls(filters)

    @property
    def filters(self) -> Sequence[Filter]:
        """A list of `Filter` instance."""
        return self._filters

    def __init__(self, filters: Optional[Sequence[Filter]] = None):
        self._filters = []
        if filters:
            for filter_ in filters:
                self.add_filter(filter_)
        assert isinstance(self._filters, Sequence)

    def __repr__(self) -> str:
        return f"<Query '{self.to_string()}'>"

    def __str__(self) -> str:
        return self.to_string()

    def add_filter(self, filter_: Filter):
        """Adds a `Filter` to the query.

        If a Filter with identical value exists in the Query, the operation
        is ignored. This guarantees that the Query has no identical Filter.
        """
        assert isinstance(
            filter_, Filter
        ), f"Expects `Filter` instance, but got {filter_}"
        self.remove_filter(filter_)
        self._filters.append(filter_)

    def remove_filter(self, filter_: Filter):
        """Removes a `Filter` to the query.

        If no Filter with identical value exists in the Query, the operation
        is ignored.
        """
        self._filters = list(filter(lambda x: x != filter_, self._filters))

    def to_string(self) -> str:
        """Return a query expression.

        >>> q = Query.parse("foo=bar foo=baz")
        >>> q.to_string()
        "foo=bar foo=baz"
        """
        return " ".join(f.to_string() for f in self.filters)

    def to_json(self) -> List[Dict]:
        """Return a list of filter dictionaries."""
        return [f.to_json() for f in self.filters]


class _Visitor(QueryExprParserVisitor):
    """Private implementation of the visitor pattern.

    Walks the parser tree and generate the filters.
    """

    def visitExprs(self, ctx: QueryExprParser.ExprsContext):
        filters = []
        self._filters = filters
        self.visitChildren(ctx)
        self._filters = None
        return filters

    def visitExpr(self, ctx: QueryExprParser.ExprContext):
        identifier = ctx.IDENTIFIER().getText()
        operator = ctx.OPERATOR().getText()
        value: Any = ""
        if ctx.NULL() is not None:
            value = None
        elif ctx.STRING() is not None:
            value = ctx.STRING().getText()
            if is_quoted_string(value):
                # Edge case where value is quoted but has no spaces
                # this gets pass the parser. eg '"NULL"'
                value = trim(value, '"', '"')
        elif ctx.STRING_QUOTED() is not None:
            value = json.loads(ctx.STRING_QUOTED().getText())
        elif ctx.BOOL() is not None:
            value = json.loads(ctx.BOOL().getText())
        elif ctx.NUMBER() is not None:
            value = ctx.NUMBER().getText()
            value = int(value) if value.lstrip("-+").isdigit() else float(value)
        label = None
        if ctx.LABEL() is not None:
            label = ctx.LABEL().getText()
            label = trim(label, "[", "]")
            if label.startswith('"') and label.endswith('"'):
                label = json.loads(label)
        self._filters.append(Filter(identifier, operator, value, label))


def trim(value: str, left: Optional[str], right: Optional[str]) -> str:
    """Removes `left` and `right` characters from `value` if they exists."""
    if not value:
        return value
    if value.startswith(left):
        value = value[len(left) :]
    if value.endswith(right):
        value = value[0 : -len(right)]
    return value


def quote_string(value: str) -> str:
    """Wrap `value` in double quotes, escaping any double quotes
    in `value` if required.
    """
    return json.dumps(value)


def unquote_string(value: str) -> str:
    """Unwrap `value` from double quotes, unescaping any double quotes
    in `value` if required.
    """
    try:
        return json.loads(value)
    except Exception:
        return value


def is_quoted_string(value: str) -> bool:
    """Returns True if `value` is wrapped in double quotes."""
    return isinstance(value, str) and value.startswith('"') and value.endswith('"')


def safe_string(value: str) -> str:
    """Wraps a string in double quotes if required, so that it can be safely
    embeded in a query expression string.
    """
    # We wrap `value if:
    # - it contains spaces, for `Filter.value`
    # - it contains square brackets, for `Filter.label`
    if isinstance(value, str) and any(c in value for c in " []"):
        return quote_string(value)
    return value



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['tgr_wdn_query_lang', 'tgr_wdn_query_lang.generated']

package_data = \
{'': ['*']}

install_requires = \
['antlr4-python3-runtime>=4.9.2,<5.0.0']

setup_kwargs = {
    'name': 'tgr-wdn-query-lang',
    'version': '1.1.5',
    'description': 'A library for parsing simple query expression.',
    'long_description': 'None',
    'author': 'Ivan Choo',
    'author_email': 'ivan@horangi.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
